;BLINKEN-LEDs PLUS     (C) ARNE ROSSIUS
;ATMega16 @ 14.7456MHz XTAL, Fuses: low = 0x1f, high = 0xca
;USE "AVRA" FOR ASSEMBLY, OTHER ASSEMBLERS MIGHT FAIL!
;
;PORTS:
;  PA0-7 = LED Col 8..1 (MSB left)
;  PB0-7 = LED Row 1..8 (MSB bottom)
;  PC0-7 = LED Col 16..9 (MSB left)
;  PD0/1 = RS232 Rx/Tx
;  PD2   = SD-Card #CS
;  PD3   = SD-Card D_in (MOSI)
;  PD4   = SD-Card D_out (MISO)
;  PD5   = SD-Card Clock
;  PD6/7 = LED Col 18/17 (MSB left) / Baud Rate jumpers (1k to GND)

;Stream Timeout: duration after which to return to flash animation if
;                no valid data was received for the entire duration
;                (set to zero to disable timeout)
.equ	STREAM_TIMEOUT_VAL = 30

;Enable SD card functionality
.equ	USE_SDCARD = 1

;Blinkstroem Advanced compatibility for BS2.BIN file playback:
;  0 = actual speed (50 Hz ticks)
;  1 = Blinkstroem Advanced speed (approx. 62 Hz ticks)
.equ	BLINKSTROEM_COMPATIBLE = 0

;Enable Stream functionalities:
;  STREAM_INPUT:  receive and display MCUF stream when no SD Card is inserted
;  STREAM_OUTPUT: output an MCUF stream of the display contents
.equ	STREAM_INPUT = 1
.equ	STREAM_OUTPUT = 0

.include "../include/m16def.inc"

;R1:R0 = 'mul' result
.def	size1 = R2
.def	size2 = R3
.def	size3 = R4
.def	size4 = R5
.def	sectorL = R6
.def	sectorH = R7
.def	framelen = R8
.def	rowlen = R9
.def	base1 = R10
.def	base2 = R11
.def	base3 = R12
.def	base4 = R13
.def	txpos = R14
.def	bits = R15

.def	temp = R16
.def	temp2 = R17
.def	count = R18
.def	pwm = R19
.def	timeL = R20
.def	timeH = R21
.def	row = R22
.def	flags = R23
	.equ	fFlashAnimation = 0 ;flash animation enabled
	.equ	fActiveFrame = 1 ;active frame (currently being displayed), 0 or 1
	.equ	fBS2BIN = 2 ;BS2.BIN file found
	.equ	fSDHC = 3 ;SDHC card (uses sector address, not byte address)
.def	Z3 = R24 ;additional bytes for 32 bit Z pointer (SD card address)
.def	Z4 = R25
;R27:R26 = X pointer (SD card data length)
;R29:R28 = Y pointer (main program RAM pointer)
;R31:R30 = Z pointer (static Animation Flash pointer)
.def	Z1 = R30
.def	Z2 = R31

.equ	SD_CS = 2
.equ	SD_MOSI = 3
.equ	SD_MISO = 4
.equ	SD_CK = 5

;===============================================================================

.macro breqx
	;branch if equal, long distance
	brne	PC+2
	rjmp	@0
.endmacro

.macro brnex
	;branch if not equal, long distance
	breq	PC+2
	rjmp	@0
.endmacro

.macro ldiw
	;load immediate word
	ldi	@0L, LOW(@1)
	ldi	@0H, HIGH(@1)
.endmacro

.macro ldid
	;load immediate double-word
	ldi	@01, BYTE1(@1)
	ldi	@02, BYTE2(@1)
	ldi	@03, BYTE3(@1)
	ldi	@04, BYTE4(@1)
.endmacro

.macro movd
	;move double-word
	movw	@02:@01, @12:@11
	movw	@04:@03, @14:@13
.endmacro

.macro subw
	;subtract immediate word from lower registers
	ldi	temp, LOW(@1)
	sub	@0L, temp
	ldi	temp, HIGH(@1)
	sbc	@0H, temp
.endmacro

.macro ldsw
	;load from immediate RAM address word
	lds	@0L, @1
	lds	@0H, @1+1
.endmacro

.macro ldsd
	;load from immediate RAM address double-word
	lds	@01, @1
	lds	@02, @1+1
	lds	@03, @1+2
	lds	@04, @1+3
.endmacro

.macro lddw
	;load from indexed RAM pointer word
	ldd	@0L, @1
	ldd	@0H, @1+1
.endmacro

.macro lddd
	;load from indexed RAM pointer double-word
	ldd	@01, @1
	ldd	@02, @1+1
	ldd	@03, @1+2
	ldd	@04, @1+3
.endmacro

.macro stsw
	;store to immediate RAM address word
	sts	@0, @1L
	sts	@0+1, @1H
.endmacro

.macro stsd
	;store to immediate RAM address double-word
	sts	@0, @11
	sts	@0+1, @12
	sts	@0+2, @13
	sts	@0+3, @14
.endmacro

.macro pushw
	;push word
	push	@0L
	push	@0H
.endmacro

.macro pushd
	;push double-word
	push	@01
	push	@02
	push	@03
	push	@04
.endmacro

.macro popw
	;pop word
	pop	@0H
	pop	@0L
.endmacro

.macro popd
	;pop double-word
	pop	@04
	pop	@03
	pop	@02
	pop	@01
.endmacro

.macro zero_w
	;compare word to zero
	cpi	@0L, 0
	cpc	@0H, @0L
.endmacro

.macro zero_d
	;compare double-word to zero
	clc
	tst	@01
	cpc	@02, @01
	cpc	@03, @01
	cpc	@04, @01
.endmacro

.macro activeframe
	;load active frame address
	ldiw	@0, RAM_Frame0
	sbrs	flags, fActiveFrame
	rjmp	activeframe_end
	ldiw	@0, RAM_Frame1
activeframe_end:
.endmacro

.macro inactiveframe
	;load inactive (buffer) frame address
	ldiw	@0, RAM_Frame1
	sbrs	flags, fActiveFrame
	rjmp	inactiveframe_end
	ldiw	@0, RAM_Frame0
inactiveframe_end:
.endmacro

.macro newframe
	;display new frame and send new frame to UART
	ldi	temp, 1<<fActiveFrame ;swap active and inactive frame
	eor	flags, temp
	.if (STREAM_OUTPUT != 0)
	sbi	UCSRB, UDRIE ;enable UDR empty interrupt
	.endif
.endmacro

.macro file_open
	;"open" file from directory
	;backup all pointers of current position in directory
	push	count
	pushw	sector
	pushw	Y
	pushd	Z
	ldsw	Z, RAM_FAT_Cluster
	pushw	Z
	;load values for file into FAT registers
	lddw	X, Y+0x1a ;first cluster of file
	stsw	RAM_FAT_Cluster, X
	lddd	size, Y+0x1c ;file size
.endmacro

.macro file_close
	;"close" file (continue reading directory index)
	;restore directory pointers
	popw	Z
	stsw	RAM_FAT_Cluster, Z
	popd	Z
	;re-load last used sector
	sbrc	flags, fSDHC
	rjmp	file_close_sdhc
	;SD: subtract 0x200 from byte address
	subi	Z2, 2
	sbci	Z3, 0
	sbci	Z4, 0
	rjmp	file_close_load
file_close_sdhc:
	;SDHC: subtract 1 from sector address
	subi	Z1, 1
	sbci	Z2, 0
	sbci	Z3, 0
	sbci	Z4, 0
file_close_load:
	cbi	PORTD, SD_CS
	rcall	sd_readblock
	sbi	PORTD, SD_CS
	popw	Y
	popw	sector
	pop	count
	
.endmacro

;===============================================================================

.dseg
	RAM_Frame0: .byte 144 ;one frame, 18x8 pixels
	RAM_Frame1: .byte 144 ;another frame, 18x8 pixels
	RAM_Duration: .byte 2 ;duration of next frame
	RAM_Block: .byte 512 ;block from SD card
	RAM_Line: .byte 80 ;one line of text
	RAM_String: .byte 10 ;one short line of text (for string comparision)
	RAM_FAT_Start: .byte 4 ;start address of first FAT
	RAM_FAT_Clustersize: .byte 1 ;sectors per cluster
	RAM_FAT_Cluster: .byte 2 ;current cluster number
	RAM_FAT_Filestart: .byte 2 ;first cluster of file (used when rewinding file/dir)
	RAM_FAT_Filesize: .byte 4 ;backup of file size (used when rewinding file)
	
	RAM_Timeout_Stream: .byte 1 ;timeout counter for Stream display
	RAM_Timeout_RS232: .byte 1 ;timeout counter for RS232 reception

	RAM_Stack: .byte 96

;===============================================================================

.cseg
.org 0x000
	rjmp	reset
	
.org OC2addr
	rjmp	oc2
	
.if (STREAM_OUTPUT != 0)
.org UDREaddr
	rjmp	uart_tx
.endif

.org OC0addr
	rjmp	oc0
	
;===============================================================================

.if (USE_SDCARD != 0)
charset:
	;8x8 pixel charset, chars 0x20 to 0x7E, 8 bytes per char:
	;bytes = columns (left to right), MSB = bottom pixel
	.include "charset8x8.asm"
.endif

;===============================================================================

;Example PWM durations:
;0, 2, 4, 7, 11, 19, 35, 64
;0, 1, 5, 12, 21, 33, 47, 64
;0, 1, 2, 4, 8, 16, 32, 64
pwm_duration:
	.equ	PWMVAL_0 = 0
	.equ	PWMVAL_1 = 2
	.equ	PWMVAL_2 = 4
	.equ	PWMVAL_3 = 7
	.equ	PWMVAL_4 = 11
	.equ	PWMVAL_5 = 19
	.equ	PWMVAL_6 = 35
	.equ	PWMVAL_7 = 64
	
	;(PWM value 0 doesn't exist [LED off])
	.db 0,                   PWMVAL_1-PWMVAL_0-1
	.db PWMVAL_2-PWMVAL_1-1, PWMVAL_3-PWMVAL_2-1
	.db PWMVAL_4-PWMVAL_3-1, PWMVAL_5-PWMVAL_4-1
	.db PWMVAL_6-PWMVAL_5-1, PWMVAL_7-PWMVAL_6-1
	
mcuf_header:
	;MCUF header for transmitted frames
	.db 0x23, 0x54, 0x26, 0x66 ;magic
	.db 0x00, 0x08, 0x00, 0x12 ;height=8, width=18
	.db 0x00, 0x01, 0x00, 0x07 ;1 channel, maxval 0x07 (8 grayscales)

;===============================================================================

.if (STREAM_OUTPUT != 0)
uart_tx: ;UART data register empty interrupt => send next byte
	cbi	UCSRB, UDRIE ;disable UDR empty interrupt
	;sei <== TODO: does this work?
	push	temp
	pushw	Z
	in	temp, SREG
	push	temp
	
	ldi	temp, 12 ;header length
	cp	txpos, temp
	brsh	uart_tx_frame
	
	;send header
	ldiw	Z, mcuf_header*2
	ldi	temp, 0
	add	ZL, txpos
	adc	ZH, temp
	lpm	temp, Z
	out	UDR, temp
	rjmp	uart_tx_end
	
uart_tx_frame:
	;send frame data
	activeframe Z
	sbiw	ZH:ZL, 12 ;txpos=12 => frame offset 0
	ldi	temp, 0
	add	ZL, txpos
	adc	ZH, temp
	ld	temp, Z
	out	UDR, temp
	
uart_tx_end:
	inc	txpos
	ldi	temp, 12+144
	cp	txpos, temp
	brsh	PC+2
	sbi	UCSRB, UDRIE ;re-enable UDR empty interrupt
	brlo	PC+2
	clr	txpos
	
;uart_tx_return:
	pop	temp
	out	SREG, temp
	popw	Z
	pop	temp
	reti
.endif

;===============================================================================

oc0:	;Timer 0 output compare interrupt (100 Hz ticks)
	sei
	push	temp
	in	temp, SREG
	push	temp
	
	;decrement frame display duration if > 0
	zero_w	time
	breq	oc0_time_end
	subi	timeL, LOW(1)
	sbci	timeH, HIGH(1)
oc0_time_end:
	
.if (STREAM_INPUT != 0)
	;RS232 timeout
	lds	temp, RAM_Timeout_RS232
	tst	temp
	breq	oc0_end
	dec	temp
	sts	RAM_Timeout_RS232, temp
.endif

oc0_end:
	;return
	pop	temp
	out	SREG, temp
	pop	temp
	reti

;===============================================================================

oc2:	;Timer 2 output compare interrupt (PWM steps)
	push	temp
	push	temp2
	pushw	Z
	push	R0
	push	R1
	in	temp, SREG
	push	temp
	
	;increment PWM value
	inc	pwm
	cpi	pwm, 8
	brlo	oc2_pwminc_end ;pwm < 8
	ldi	pwm, 1
	inc	row
	andi	row, 0x07
oc2_pwminc_end:
	
	;determine next interval
	ldiw	Z, pwm_duration*2
	ldi	temp, 0
	add	ZL, pwm
	adc	ZH, temp
	lpm	temp, Z
	out	OCR2, temp
	
	;set LEDs
	activeframe Z
	ldi	temp, 18 ;1 row
	mul	temp, row
	add	ZL, R0
	adc	ZH, R1
	
	;LEDs 1 to 8
	clr	R0
	ldi	temp2, 0x80
oc2_leds1_loop:
	ld	temp, Z+
	cp	temp, pwm
	brlo	PC+2
	or	R0, temp2
	lsr	temp2
	brne	oc2_leds1_loop
	
	;LEDs 9 to 16
	clr	R1
	ldi	temp2, 0x80
oc2_leds2_loop:
	ld	temp, Z+
	cp	temp, pwm
	brlo	PC+2
	or	R1, temp2
	lsr	temp2
	brne	oc2_leds2_loop
	
	;LEDs 17, 18
	clr	temp2
	ld	temp, Z+
	cp	temp, pwm
	brlo	PC+2
	ori	temp2, 0x80
	ld	temp, Z+
	cp	temp, pwm
	brlo	PC+2
	ori	temp2, 0x40
	
	in	ZH, PORTD
	andi	ZH, 0x3F
	;switch row?
	cpi	pwm, 1
	brne	oc2_row_end
	ldi	temp, 0x01 ;generate new row bitmask
	sbrc	row, 0
	lsl	temp
	sbrc	row, 1
	lsl	temp
	sbrc	row, 1
	lsl	temp
	sbrc	row, 2
	swap	temp
	ldi	ZL, 0x00 ;LEDs off
	out	PORTA, ZL
	out	PORTC, ZL
	out	PORTD, ZH
	out	PORTB, temp ;switch row
	ldi	temp, 5 ;short delay (wait until transistors have switched)
oc2_row_delay:
	dec	temp
	brne	oc2_row_delay
oc2_row_end:
	;output new LED values
	or	ZH, temp2 ;LEDs on
	out	PORTA, R0
	out	PORTC, R1
	out	PORTD, ZH
	
	;return
	pop	temp
	out	SREG, temp
	pop	R1
	pop	R0
	popw	Z
	pop	temp2
	pop	temp
	reti
	
;===============================================================================

reset:
	;set stackpointer
	ldiw	Z, RAMEND
	out	SPH, ZH
	out	SPL, ZL
	
	;init ports
	ldi	temp, 0xFF
	out	DDRA, temp
	out	DDRB, temp
	out	DDRC, temp
.if (STREAM_INPUT != 0)
	ldi	temp, 0x2E ;PD6/7 are inputs at first (for Baud Rate jumpers)
.else
	ldi	temp, 0xEE
.endif
	out	DDRD, temp
	ldi	temp, 0x00
	out	PORTA, temp
	out	PORTB, temp
	out	PORTC, temp
	ldi	temp, 0xC3 | 1<<SD_CS | 1<<SD_MOSI
	out	PORTD, temp
	
	;init timer 0 (100 Hz ticks)
	ldi	temp, 1<<WGM01 | 0x05 ;CTC mode, Clk/1024
	out	TCCR0, temp
	ldi	temp, 144-1 ;14.7456 MHz / 1024 / 144 = 100 Hz
	out	OCR0, temp
	
	;init timer 2 (PWM)
	ldi	temp, 1<<WGM21 | 0x06 ;CTC mode, Clk/256
	out	TCCR2, temp
	
	;enable timer interrupts
	ldi	temp, 1<<OCIE2 | 1<<OCIE0
	out	TIMSK, temp
	
	;init registers
	clr	row
	clr	pwm
	clr	flags
	clr	txpos
	
	;init delay (~1 ms)
	ldi	temp, 20
	ldi	temp2, 0
init_delay:
	dec	temp2
	brne	init_delay
	dec	temp
	brne	init_delay
	
.if (STREAM_INPUT != 0)
	;init UART
	ldi	temp, 1<<RXEN | 1<<TXEN
	out	UCSRB, temp
	ldi	temp, 0
	out	UBRRH, temp
	in	temp2, PIND
	andi	temp2, 0xC0
	ldi	temp, 15 ;57600 Baud @ 14.7456 MHz
	cpi	temp2, 0x80
	brne	PC+2
	ldi	temp, 7 ;115200 Baud @ 14.7456 MHz
	cpi	temp2, 0x40
	brne	PC+2
	ldi	temp, 3 ;230400 Baud @ 14.7456 MHz
	cpi	temp2, 0x00
	brne	PC+2
	ldi	temp, 1 ;460800 Baud @ 14.7456 MHz
	out	UBRRL, temp
	ldi	temp, 0x03 | 1<<SD_CS | 1<<SD_MOSI
	out	PORTD, temp
	ldi	temp, 0xEE
	out	DDRD, temp
.endif
	
	;enable interrupts
	sei
	
	;display splash screen
	ldiw	Z, initframe*2
	rcall	frame_load
splash_wait:
	zero_w	time
	brne	splash_wait
	
;-------------------------------------------------------------------------------

;if SD card support is enabled, check if an SD card is present,
;otherwise fall through to flash animation
.if (USE_SDCARD != 0)
	rjmp	sdcard
.endif

;===============================================================================

flash:
	;display animation from flash
	ldiw	Z, frames*2
	ldiw	time, 0

flash_continue:
	ori	flags, 1<<fFlashAnimation

flash_wait:
	;wait until current frame time has elapsed
.if (STREAM_INPUT != 0)
	sbic	UCSRA, RXC ;jump to stream mode if UART received something
	rjmp	stream
.endif
	zero_w	time
	brne	flash_wait
	
	rcall	frame_load
	rjmp	flash_wait
	
;--------------------

frame_load:
	;load and display next frame from flash (Z)
	lpm	YL, Z+
	lpm	YH, Z+
	zero_w	Y ;0x0000 = end marker
	breq	frame_loadfirst
	movw	timeH:timeL, YH:YL
	inactiveframe Y
	ldi	count, 18
frame_load_loop:
	lpm	R0, Z+
	ldi	temp2, 8
frame_load_bitloop:
	ldi	temp, 0x00 ;LED off
	sbrc	R0, 0
	ldi	temp, 0x07 ;max. brightness
	lsr	R0
	st	Y, temp
	adiw	YH:YL, 18
	dec	temp2
	brne	frame_load_bitloop
	subi	YL, LOW(144 - 1)
	sbci	YH, HIGH(144 - 1)
	dec	count
	brne	frame_load_loop
	newframe
	ret
	
frame_loadfirst:
	;end of animation reached, start over
	ldiw	Z, frames*2
	rjmp	frame_load

;===============================================================================

.if (STREAM_INPUT != 0)

stream:
	;RS232 Stream
stream_framestart:
	;receive header
	ldi	temp, 100 ;timeout 1 s
	sts	RAM_Timeout_RS232, temp
	rcall	uart_getc
	cpi	temp, 0x23 ;magic byte #1
	brne	stream_error
	rcall	uart_getc
	cpi	temp, 0x54 ;magic byte #2
	brne	stream_error
	rcall	uart_getc
	cpi	temp, 0x26 ;magic byte #3
	brne	stream_error
	rcall	uart_getc
	cpi	temp, 0x66 ;magic byte #4
	brne	stream_error
	rcall	uart_getc
	cpi	temp, 0 ;height MSB
	brne	stream_error
	rcall	uart_getc
	cpi	temp, 8 ;height LSB
	brne	stream_error
	rcall	uart_getc
	cpi	temp, 0 ;width MSB
	brne	stream_error
	rcall	uart_getc
	cpi	temp, 18 ;width LSB
	brne	stream_error
	rcall	uart_getc
	cpi	temp, 0x00 ;channels MSB
	brne	stream_error
	rcall	uart_getc
	cpi	temp, 0x01 ;channels LSB
	brne	stream_error
	rcall	uart_getc
	cpi	temp, 0x00 ;maxval MSB
	brne	stream_error
	rcall	uart_getc
	cpi	temp, 0 ;maxval LSB: must be a power of two minus one (range 1 to 255)
	breq	stream_error
	mov	temp2, temp
	ldi	count, -4
stream_maxval_check:
	inc	count
	lsr	temp2
	brcs	stream_maxval_check
	cpi	temp2, 0
	breq	stream_data
	
stream_rs232_timeout:
	;timeout waiting for data
stream_error:
	;received invalid data
	sbrs	flags, fFlashAnimation
	rjmp	stream_framestart ;a valid frame was received before: wait for next frame
	rjmp	flash_continue ;no valid frame received yet: return to flash animation mode
	
stream_data:
	;receive data
	inactiveframe Y
	ldi	temp2, 144
stream_receive:
	rcall	uart_getc
	cpi	count, 0 ;count zero: don't shift
	breq	stream_store
	mov	R0, count
	sbrc	R0, 7
	rjmp	stream_shift_left
stream_shift_right: ;count positive: shift right
	lsr	temp
	dec	R0
	brne	stream_shift_right
	rjmp	stream_store
stream_shift_left: ;count negative: shift left
	lsl	temp
	inc	R0
	brne	stream_shift_left
stream_store:
	andi	temp, 0x07
	st	Y+, temp
	dec	temp2
	brne	stream_receive
stream_complete:
	;display new frame
	andi	flags, ~(1<<fFlashAnimation) ;disable flash animation
	newframe
	;done, wait for next frame
	rjmp	stream_framestart
	
;--------------------

uart_getc:
uart_getc_wait:
	lds	temp, RAM_Timeout_RS232
	tst	temp
	breq	uart_getc_timeout
	sbis	UCSRA, RXC
	rjmp	uart_getc_wait
	ldi	temp, 10 ;timeout 100 ms
	sts	RAM_Timeout_RS232, temp
	ldi	temp, STREAM_TIMEOUT_VAL
	sts	RAM_Timeout_Stream, temp
	in	temp, UDR
	ret
	
uart_getc_timeout:
	;timeout while waiting for RS232 reception
	pop	temp ;remove return address from stack
	pop	temp
	lds	temp, RAM_Timeout_Stream
	tst	temp
	breqx	flash_continue
	dec	temp
	sts	RAM_Timeout_Stream, temp
	rjmp	stream_rs232_timeout
	
.endif
	
;===============================================================================

.if (USE_SDCARD != 0)

sdcard:
	rcall	sd_init
	cpi	temp, 0
	brnex	flash ;play animation from flash if SD init failed
	
	;read MBR, find FAT16 partition
	cbi	PORTD, SD_CS
	ldid	Z, 0
	rcall	sd_readblock
	ldsw	X, RAM_Block+0x1FE
	cpi	XL, 0x55
	brne	sd_mbr_error
	cpi	XH, 0xAA
	brne	sd_mbr_error
	ldi	count, 4
	ldiw	Y, RAM_Block+0x1BE
sd_partition:
	ldd	temp, Y+4 ;read partition type
	cpi	temp, 6 ;FAT16
	breq	sd_partition_found
	cpi	temp, 4 ;FAT16 <32MB
	breq	sd_partition_found
	adiw	YH:YL, 16
	dec	count
	brne	sd_partition
	
	;no FAT16 partition found
	sbi	PORTD, SD_CS
	rcall	message
		.db "No FAT16 Partition found.",0
	rjmp	flash
	
sd_mbr_error:
	;faulty MBR or no FAT16 partition found
	sbi	PORTD, SD_CS
	rcall	message
		.db "Invalid MBR!",0
	rjmp	flash
	
sd_partition_found:
	;read partition data
	sbrc	flags, fSDHC
	rjmp	sd_partition_sdhc
	;SD: convert sector address to byte address
	clr	Z1
	ldd	Z2, Y+8 ;LBA of first sector, byte 1
	ldd	Z3, Y+9 ;LBA of first sector, byte 2
	ldd	Z4, Y+10 ;LBA of first sector, byte 3
	lsl	Z2
	rol	Z3
	rol	Z4
	rjmp	sd_partition_end
sd_partition_sdhc:
	;SDHC
	lddd	Z, Y+8 ;LBA of first sector
sd_partition_end:

	;read FAT16 boot sector
	rcall	sd_readblock
	
	lds	temp, RAM_Block+0xD ;sectors per cluster
	sts	RAM_FAT_Clustersize, temp
	
	ldsw	X, RAM_Block+0xE ;reserved sector count => start of first FAT
	sbiw	XH:XL, 1 ;first sector already read
	sbrc	flags, fSDHC
	rjmp	sd_fatstart_sdhc
	;SD: convert sector address to byte address
	clr	count
	lsl	XL
	rol	XH
	rol	count
	add	Z2, XL
	adc	Z3, XH
	adc	Z4, count
	rjmp	sd_fatstart_end
sd_fatstart_sdhc:
	;SDHC
	ldi	temp, 0
	add	Z1, XL
	adc	Z2, XH
	adc	Z3, temp
	adc	Z4, temp
sd_fatstart_end:
	stsd	RAM_FAT_Start, Z
	
	;calculate root directory address (directly after the last FAT)
	ldsd	base, RAM_FAT_Start ;start address of first FAT
	ldsw	X, RAM_Block+0x16 ;sectors per FAT
	lds	temp, RAM_Block+0x10 ;number of FATs
	ldi	temp2, 0
	sbrc	flags, fSDHC
	rjmp	sd_firstcluster_sdhc
	;SD: multiply sectors/FAT * FATs, convert result to byte offset
	clr	count
	lsl	XL
	rol	XH
	rol	count
	mul	XL, temp
	add	base2, R0
	adc	base3, R1
	adc	base4, temp2
	mul	XH, temp
	add	base3, R0
	adc	base4, R1
	mul	temp, count
	add	base4, R0
	rjmp	sd_firstcluster_end
sd_firstcluster_sdhc:
	;SDHC: multiply sectors/FAT * FATs (result is sector offset)
	mul	XL, temp
	add	base1, R0
	adc	base2, R1
	adc	base3, temp2
	adc	base3, temp2
	mul	XH, temp
	add	base2, R0
	adc	base3, R1
	adc	base4, temp2
sd_firstcluster_end:
	
	;move root directory address to Z, add root directory length to 'base'
	movd	Z, base
	ldsw	Y, RAM_Block+0x11 ;number of root directory entries
	ldi	temp2, 0
	sbrc	flags, fSDHC
	rjmp	sd_rootdir_sdhc
	;SD
	ldi	temp, 32 ;32 bytes per directory entry
	mul	YL, temp
	add	base1, R0
	adc	base2, R1
	adc	base3, temp2
	adc	base4, temp2
	mul	YH, temp
	add	base2, R0
	adc	base3, R1
	adc	base4, temp2
	rjmp	sd_rootdir_end
sd_rootdir_sdhc:
	;SDHC: root directory length must be a multiple of 512 bytes
	mov	temp, YL
	andi	temp, 0x0F ;16 entries/sector => lower 4 bits must be 0
	brnex	sd_rootdir_error
	andi	YL, 0xF0 ;Y >>= 4 (# of entries / 16 = root directory sectors)
	swap	YL
	mov	temp, YH
	andi	temp, 0x0F
	swap	temp
	or	YL, temp
	andi	YH, 0xF0
	swap	YH
	ldi	temp, 0
	add	base1, YL
	adc	base2, YH
	adc	base3, temp
	adc	base4, temp
sd_rootdir_end:
	
	
	
sd_scan_root:
	;find directory "BLPLUS" or file BS2.BIN
	rcall	sd_readblock
	ldiw	Y, RAM_Block
	ldi	count, 512/32 ;512 bytes/sector, 32 bytes/entry
sd_scan_root_entry:
	ldd	temp, Y+0xB ;file attributes
	cpi	temp, 0x0F ;ignore LFN entries
	breqx	sd_scan_root_next
	ld	temp, Y
	cpi	temp, 0x00 ;no more entries
	breqx	sd_scan_root_end
	cpi	temp, 0xE5 ;deleted file
	breq	sd_scan_root_next
	
	;check for "BLPLUS" directory
	movw	XH:XL, YH:YL
	rcall	str_compare
		.db "BLPLUS     ",0
	brne	sd_scan_root_blplus_end
	ldd	temp, Y+0xB ;file attributes
	sbrs	temp, 4 ;must be a directory
	rjmp	sd_scan_root_blplus_end
	andi	temp, ~(0x37) ;allowed attributes: readonly, hidden, system, directory, archive
	brne	sd_scan_root_blplus_end
	sbi	PORTD, SD_CS
	rjmp	sd_blplus
sd_scan_root_blplus_end:
	
	;check for "BS2.BIN" file
	movw	XH:XL, YH:YL
	rcall	str_compare
		.db "BS2     BIN",0
	brne	sd_scan_root_bs2bin_end
	ldd	temp, Y+0xB ;file attributes
	andi	temp, ~(0x27) ;allowed attributes: readonly, hidden, system, archive
	brne	sd_scan_root_blplus_end
	ori	flags, 1<<fBS2BIN
	lddd	size, Y+0x1C ;read file size
	zero_d	size
	breq	sd_scan_root_bs2bin_end ;ignore empty file
	stsd	RAM_FAT_Filesize, size
	lddw	X, Y+0x1a ;copy first cluster address to RAM_FAT_Filestart
	stsw	RAM_FAT_Filestart, X
sd_scan_root_bs2bin_end:

sd_scan_root_next:
	;next entry
	adiw	YH:YL, 32
	cp	Z1, base1
	cpc	Z2, base2
	cpc	Z3, base3
	cpc	Z4, base4
	brsh	sd_scan_root_end
	dec	count
	brnex	sd_scan_root_entry
	rjmp	sd_scan_root

sd_scan_root_end:
	;no "BLPLUS" directory found, use BS2.BIN file if present
	sbi	PORTD, SD_CS
	sbrc	flags, fBS2BIN
	rjmp	sd_bs2bin
	;SD card contains no usable data
	rcall	message
		.db "No usable files found, sleeping...",0
	napf: rjmp napf
	
sd_rootdir_error:
	;root directory length on SDHC card is not a multiple of 512 bytes
	sbi	PORTD, SD_CS
	rcall	putstr
		.db "SDHC Root Dir error.",0
	rjmp	flash
	
;--------------------

sd_bs2bin:
	rcall	putstr
		.db "Found BS2.BIN file",0x0a,0
	rcall	message
		.db "BS2.BIN",0
	
	;reconfigute 100 Hz timer for selected playback speed
	;(division by 2 happens later by multiplying frame durations by 2)
	;speed according to converter scripts: 50 Hz
	;actual Blinkstroem advanced speed: ~62 Hz
	.if (BLINKSTROEM_COMPATIBLE == 0)
	ldi	temp, 144-1 ;14.7456 MHz / 1024 / 144 / 2 = 50 Hz
	.else
	ldi	temp, 116-1 ;14.7456 MHz / 1024 / 116 / 2 = ~62 Hz
	.endif
	out	OCR0, temp
	
	ldiw	time, 0
sd_bs2bin_loop:
	ldsw	X, RAM_FAT_Filestart
	stsw	RAM_FAT_Cluster, X
	ldsd	size, RAM_FAT_Filesize
	rcall	sd_bin
	rjmp	sd_bs2bin_loop
	
;--------------------

sd_blplus:
	sbi	PORTD, SD_CS
	lddw	X, Y+0x1a ;copy first directory cluster address to RAM_FAT_Filestart
	stsw	RAM_FAT_Filestart, X
	rcall	message
		.db "BLPLUS",0
	
sd_blplus_loop:
	ldsw	X, RAM_FAT_Filestart
	stsw	RAM_FAT_Cluster, X
	rcall	sd_read_first_sector
sd_blplus_dir:
	ld	temp, Y
	cpi	temp, 0xE5 ;deleted file
	breq	sd_blplus_next
	cpi	temp, 0x00 ;end of directory
	breq	sd_blplus_eod
	ldd	temp, Y+0xB ;file attributes
	andi	temp, ~(0x27) ;allowed attributes: readonly, hidden, system, archive
	brne	sd_blplus_next
	;check file type
	pushw	Y
	ldi	temp2, 11
sd_blplus_name:
	ld	temp, Y+
	rcall	putc
	cpi	temp2, 4
	brne	PC+3
	ldi temp, ' '
	rcall putc
	dec	temp2
	brne	sd_blplus_name
	popw	Y
	
	stsd	RAM_FAT_Filesize, size
	movw	XH:XL, YH:YL
	adiw	XH:XL, 8
	rcall	str_compare
		.db "BIN",0 ;BS2.BIN-format (Blinkstroem) file
	breq	sd_blplus_bin
	movw	XH:XL, YH:YL
	adiw	XH:XL, 8
	rcall	str_compare
		.db "BLM",0 ;BLM (BlinkenLights Movie) file
	breqx	sd_blplus_blm
	movw	XH:XL, YH:YL
	adiw	XH:XL, 8
	rcall	str_compare
		.db "BML",0 ;BML (Blinkenlights Markup Language) file
	breqx	sd_blplus_bml
	ldi temp, 0x0a
	rcall putc
sd_blplus_next:
	adiw	YH:YL, 32
	subw	sector, 32
	breq	sd_blplus_sector
	rjmp	sd_blplus_dir
sd_blplus_sector:
	rcall	sd_read_sector
	zero_w	Y
	breq	sd_blplus_eod
	rjmp	sd_blplus_dir
	
sd_blplus_eod:
	rjmp	sd_blplus_loop
	
;--------------------

sd_blplus_bin:
	file_open
	rcall putstr
		.db "  ==> BIN",0x0a,0
	rcall	sd_bin
	file_close
	rjmp	sd_blplus_next

;--------------------
	
sd_blplus_blm:
	rcall putstr
		.db "  ==> BLM",0x0a,0
	rjmp	sd_blplus_next

;--------------------
	
sd_blplus_bml:
	file_open
	rcall putstr
		.db "  ==> BML",0x0a,0
	rcall	sd_read_first_sector
sd_bml_start:
	rcall	sd_read_xml_tag
	zero_w	Y
	breq	sd_bml_eof
	ldiw	X, RAM_Line
	rcall	str_compare
		.db "blm ",0
	brne	sd_bml_start
	pushw	Y
	rcall	str2ram		;width == 18 ?
		.db "width=",0
	rcall	str_xml_param
	ldi	temp, 1
	cpi	temp2, 8
	cpc	XL, temp
	brne	sd_bml_invalid
	rcall	str2ram		;height == 8 ?
		.db "height=",0
	rcall	str_xml_param
	cpi	temp2, 8
	brne	sd_bml_invalid
	rcall	str2ram		;bits per pixel (valid values are 1..8)
		.db "bits=",0
	rcall	str_xml_param
	cpi	temp2, 0
	breq	sd_bml_invalid
	cpi	temp2, 9
	brsh	sd_bml_invalid
	mov	bits, temp2
	popw	Y
	
sd_bml_read:
	rcall	sd_read_xml_tag
	zero_w	Y
	breq	sd_bml_eof
	ldiw	X, RAM_Line
	rcall	str_compare
		.db "frame",0
	breq	sd_bml_frame
	rjmp	sd_bml_read
	
sd_bml_invalid:
	;invalid BML file, skip
	popw	Y
sd_bml_eof:
	;end of file
	file_close
	rjmp	sd_blplus_next

sd_bml_frame:
	pushw	Y
	rcall	str2ram
		.db "duration=",0
	rcall	str_xml_param
	popw	Y
	cpi	temp2, 0xFF
	brne	sd_bml_frame_read
	ldiw	X, 1 ;default duration: 10 ms
sd_bml_frame_read:
	stsw	RAM_Duration, X
	ldi	temp, 144
	mov	framelen, temp
sd_bml_frame_loop:
	rcall	sd_read_xml_tag
	zero_w	Y
	breq	sd_bml_eof
	ldiw	X, RAM_Line
	rcall	str_compare
		.db "/frame",0
	breqx	sd_bml_frame_end
	ldiw	X, RAM_Line
	rcall	str_compare
		.db "row"
	breq	sd_bml_frame_data
	rjmp	sd_bml_frame_loop
	
sd_bml_frame_data:
	;read one row from SD card
	tst	framelen
	breq	sd_bml_frame_loop ;ignore row if frame is already complete
	ldi	temp, 18
	mov	rowlen, temp
	ldi	temp, 5
	cp	bits, temp
	brlo	PC+2
	lsl	rowlen ;5 to 8 bits per pixel: 2 hex digits per pixel
sd_bml_data_loop:
	rcall	sd_read_byte
	zero_w	Y
	breqx	sd_bml_eof
	cpi	temp, '<'
	breq	sd_bml_data_end ;end of row data
	cpi	temp, '0'
	brlo	sd_bml_data_loop
	cpi	temp, '9'+1
	brlo	sd_bml_data_valid
	cpi	temp, 'A'
	brlo	sd_bml_data_loop
	cpi	temp, 'F'+1
	brlo	sd_bml_data_valid
	cpi	temp, 'a'
	brlo	sd_bml_data_loop
	cpi	temp, 'f'+1
	brsh	sd_bml_data_loop
sd_bml_data_valid:
	;valid byte (0..9, A..F or a..f)
	tst	rowlen
	breq	sd_bml_data_loop ;ignore byte if row is already complete
	dec	rowlen
	mov	temp2, bits
	cpi	temp2, 5
	brlo	sd_bml_data_store
	subi	temp2, 4
	sbrs	rowlen, 0 ;ignore even (lower) hex digits
	rjmp	sd_bml_data_loop
sd_bml_data_store:
	;TODO: use both bytes for 5/6 bpp
	cpi	temp, 'a'
	brlo	PC+2
	subi	temp, ('a' - 'A') ;convert to uppercase
	subi	temp, '0'
	cpi	temp, 10
	brlo	PC+2
	subi	temp, 7
	cpi	temp2, 4 ;4 (8) bits per pixel
	brne	PC+2
	lsr	temp
	cpi	temp2, 2 ;2 (6) bits per pixel
	brne	PC+4
	lsl	temp
	sbrc	temp, 2
	ori	temp, 0x01
	cpi	temp2, 1 ;1 (5) bits per pixel
	brne	PC+5
	lsl	temp
	lsl	temp
	sbrc	temp, 2
	ori	temp, 0x03
	inactiveframe X
	ldi	temp2, 144
	add	XL, temp2
	ldi	temp2, 0
	adc	XH, temp2
	sub	XL, framelen
	sbc	XH, temp2
	andi	temp, 0x07
	st	X, temp
	dec	framelen
	rjmp	sd_bml_data_loop
	
sd_bml_data_end:
	;end of row data
	tst	rowlen
	breqx	sd_bml_frame_loop
	ldi	temp, 0
	inactiveframe X
	ldi	temp2, 144
	add	XL, temp2
	ldi	temp2, 0
	adc	XH, temp2
	sub	XL, framelen
	sbc	XH, temp2
	st	X, temp
	dec	framelen
	dec	rowlen
	rjmp	sd_bml_data_end
	
sd_bml_frame_end:
sd_bml_wait:
	zero_w	time
	brne	sd_bml_wait
	ldsw	time, RAM_Duration
	newframe
	rjmp	sd_bml_read
	
;--------------------

sd_bin:
	;play binary file starting at cluster X with size RAM_FAT_Filesize
	rcall	sd_read_first_sector
sd_bin_frame:
	;read next frame
	ld	XL, Y+
	ld	XH, Y+
	lsl	XL ;multiply by 2
	rol	XH
	stsw	RAM_Duration, X
	subw	sector, 2
	brne	sd_bin_duration_end
	rcall	sd_read_sector
	zero_w	Y ;error?
	breqx	sd_bin_error
sd_bin_duration_end:
	inactiveframe X
	ldi	temp2, 144
sd_bin_pixel:
	ld	temp, Y+
	andi	temp, 0x07
	st	X, temp
	adiw	XH:XL, 18
	ld	temp, Y+
	andi	temp, 0x07
	st	X, temp
	adiw	XH:XL, 18
	subw	sector, 2
	brne	sd_bin_pixel_end
	push	temp2
	pushw	X
	rcall	sd_read_sector
	zero_w	Y ;error?
	breq	sd_bin_error
	popw	X
	pop	temp2
sd_bin_pixel_end:
	subi	temp2, 2
	mov	temp, temp2
	andi	temp, 0x07
	brne	sd_bin_pixel ;repeat 4 times (2*4 pixels/column)
	subi	XL, LOW(144 - 1)
	sbci	XH, HIGH(144 - 1)
	cpi	temp2, 0
	brne	sd_bin_pixel ;repeat 18 times (18 columns/frame)
	;wait until frame should be displayed
sd_bin_wait:
	zero_w	time
	brne	sd_bin_wait
	ldsw	time, RAM_Duration
	newframe
	;decrement remaining file size
	ldi	temp, 0
	ldiw	X, 144+2
	sub	size1, XL
	sbc	size2, XH
	sbc	size3, temp
	sbc	size4, temp
	brcs	sd_bin_eof
	breq	sd_bin_eof
	rjmp	sd_bin_frame

sd_bin_eof:
sd_bin_error:
	ret

;--------------------
	
sd_read_xml_tag:
	;read next XML tag from SD card (TODO: prevent buffer overflow: stop after 80 chars)
	zero_w	Y
	breq	sd_read_xml_tag_eof
	rcall	sd_read_byte
	cpi	temp, '<'
	brne	sd_read_xml_tag
	ldiw	X, RAM_Line
sd_read_xml_tag_loop:
	zero_w	Y
	breq	sd_read_xml_tag_eof
	rcall	sd_read_byte
	cpi	temp, '>'
	breq	sd_read_xml_tag_end
	st	X+, temp
	rjmp	sd_read_xml_tag_loop
sd_read_xml_tag_eof:
sd_read_xml_tag_end:
	ldi	temp, 0
	st	X, temp
	ret

;--------------------

sd_read_byte:
	;read 1 byte from SD card
	ld	temp2, Y+ ;read 1 byte
	ldi	temp, 1 ;decrement file size
	sub	size1, temp
	ldi	temp, 0
	sbc	size2, temp
	sbc	size3, temp
	sbc	size4, temp
	subw	sector, 1 ;modifies temp!
	brne	sd_read_byte_end
	pushw	X
	push	temp2
	rcall	sd_read_sector
	pop	temp2
	popw	X
sd_read_byte_end:
	mov	temp, temp2
	zero_d	size
	breq	sd_read_byte_eof
	ret
	
sd_read_byte_eof:
	ldiw	Y, 0
	ldi	temp, 0
	ret

;--------------------

sd_read_sector:
	;read sector from SD card (find and advance to next cluster if necessary)
	cbi	PORTD, SD_CS
	dec	count
	brnex	sd_read_sector_read ;same cluster as before
	;read next cluster number from FAT
	ldsw	X, RAM_FAT_Cluster
	ldsd	Z, RAM_FAT_Start
	ldi	temp, 0
	;calculate sector address of wanted FAT entry
	sbrc	flags, fSDHC
	rjmp	sd_read_sector_fat_sdhc
	;SD
	add	Z2, XH
	adc	Z3, temp
	adc	Z4, temp
	add	Z2, XH
	adc	Z3, temp
	adc	Z4, temp
	rjmp	sd_read_sector_fat_end
sd_read_sector_fat_sdhc:
	;SDHC
	add	Z1, XH
	adc	Z2, temp
	adc	Z3, temp
	adc	Z4, temp
sd_read_sector_fat_end:
	push	XL
	rcall	sd_readblock
	pop	XL
	ldi	YH, 0
	ldiw	Y, RAM_Block
	add	YL, XL ;add FAT entry # within sector twice (2 bytes per FAT entry)
	adc	YH, XH
	add	YL, XL
	adc	YH, XH
	ld	XL, Y+ ;get next cluster number
	ld	XH, Y+
	ldi	temp, 0
	cpi	XL, 0x02 ;cluster number must be >= 2
	cpc	XH, temp
	brlo	sd_read_sector_error
	ldi	temp, 0xFF
	cpi	XL, 0xF0 ;cluster number must be < 0xFFF0
	cpc	XH, temp
	brsh	sd_read_sector_error
	stsw	RAM_FAT_Cluster, X
sd_read_first_sector:
	;calculate start of sector address from cluster number
	cbi	PORTD, SD_CS
	ldsw	X, RAM_FAT_Cluster
	sbiw	XH:XL, 2 ;base address points to cluster 2
	movd	Z, base
	lds	count, RAM_FAT_Clustersize
	ldi	temp, 0
	sbrc	flags, fSDHC
	rjmp	sd_read_sector_addr_sdhc
	;SD
	mul	XL, count
	add	Z2, R0 ;add twice (1 sector = 2*256 bytes)
	adc	Z3, R1
	adc	Z4, temp
	add	Z2, R0
	adc	Z3, R1
	adc	Z4, temp
	mul	XH, count
	add	Z3, R0
	adc	Z4, R1
	add	Z3, R0
	adc	Z4, R1
	rjmp	sd_read_sector_addr_end
sd_read_sector_addr_sdhc:
	;SDHC
	mul	XL, count
	add	Z1, R0
	adc	Z2, R1
	adc	Z3, temp
	adc	Z4, temp
	mul	XH, count
	add	Z2, R0
	adc	Z3, R1
	adc	Z4, temp
sd_read_sector_addr_end:
sd_read_sector_read:
	;read sector
	rcall	sd_readblock
	sbi	PORTD, SD_CS
	;pre-load RAM pointer and sector word (= 2 bytes) counter
	ldiw	Y, RAM_Block
	ldi	temp, HIGH(512)
	mov	sectorH, temp
	clr	sectorL ;LOW(512) = 0
	ret
	
sd_read_sector_error:
	;read invalid cluster number (valid range is 0x0002 to 0xFFEF)
	ldiw	Y, 0 ;return RAM address zero (invalid)
	ret

;--------------------

sd_readblock:
	;read 512 byte block from SD card at address Z4:Z3:Z2:Z1 to RAM_Block,
	;increment SD card address by number of read bytes (SD)/sectors (SDHC)
	ldi	temp, 17 ;CMD17: read single block
	rcall	sd_command
	sbrc	flags, fSDHC
	rjmp	sd_readblock_sdhc
	subi	Z1, BYTE1(-512) ;SD: add 512 to Z
	sbci	Z2, BYTE2(-512)
	sbci	Z3, BYTE3(-512)
	sbci	Z4, BYTE4(-512)
	rjmp	sd_readblock_wait
sd_readblock_sdhc:
	subi	Z1, BYTE1(-1) ;SDHC: add 1 to Z
	sbci	Z2, BYTE2(-1)
	sbci	Z3, BYTE3(-1)
	sbci	Z4, BYTE4(-1)
sd_readblock_wait:
	rcall	sd_read
	cpi	temp, 0xFE
	brne	sd_readblock_wait
	ldiw	Y, RAM_Block
	ldiw	X, 512
sd_readblock_loop:
	rcall	sd_read
	st	Y+, temp
	sbiw	XH:XL, 1
	brne	sd_readblock_loop
	rcall	sd_read ;flush CRC16
	rcall	sd_read
	rcall	sd_clock
	ret

;--------------------

sd_init:
	;initalize SD card (SPI mode), return value: error code or 0 (success)
	;send at least 74 clock pulses (10 bytes = 80 pulses)
	ldi	count, 10
sd_init_pulses:
	rcall	sd_clock
	dec	count
	brne	sd_init_pulses
	
	clr	Z1
	clr	Z2
	clr	Z3
	clr	Z4
	
	;send CMD0 (software reset)
	cbi	PORTD, SD_CS
	ldi	temp, 0
	rcall	sd_command
	rcall	sd_clock
	cpi	temp, 0x01
	brnex	sd_init_error
	
	rcall putstr
	.db "CMD0 OK",0x0a,0
	
	ldid	Z, 0x122 ;2.7~3.6V, check pattern 0x22 (=> same CRC as for CMD0)
	ldi	temp, 8
	rcall	sd_command
	cpi	temp, 0x05 ;unknown command => MMC (or old SD card?)
	breqx	sd_init_mmc
	cpi	temp, 0x01
	brnex	sd_init_error
	rcall	sd_clock ;ignore byte 1
	rcall	sd_clock ;ignore byte 2
	rcall	sd_read
	cpi	temp, 0x01 ;voltage range OK?
	brnex	sd_init_error
	rcall	sd_read
	cpi	temp, 0x22 ;matches check pattern?
	brnex	sd_init_error
	rcall	sd_clock
	
	rcall putstr
	.db "CMD8 OK",0x0a,0
	
sd_init_41:
	;send ACMD41 until card leaves idle state
	ldid	Z, 0
	ldi	temp, 55
	rcall	sd_command
	rcall	sd_clock
	ldid	Z, 1<<30 ;bit 30 (HCS) set
	ldi	temp, 41
	rcall	sd_command
	rcall	sd_clock
	cpi	temp, 0x01
	breq	sd_init_41
	cpi	temp, 0x05 ;unknown command => possibly a MMC (?)
	breq	sd_init_mmc
	cpi	temp, 0x00
	brne	sd_init_error
	
	rcall putstr
	.db "ACMD41 OK",0x0a,0
	
	;send CMD58 (read OCR) to determine if card is SDHC
	ldid	Z, 0
	ldi	temp, 58
	rcall	sd_command
	cpi	temp, 0x00
	brne	sd_init_error
	rcall	sd_read
	rcall	sd_clock ;ignore byte 2
	rcall	sd_clock ;ignore byte 3
	rcall	sd_clock ;ignore byte 4
	rcall	sd_clock
	sbrs	temp, 6 ;HCS bit set?
	rjmp	sd_init_done
	
	rcall putstr
	.db "SDHC Card detected.",0x0a,0
	ori	flags, 1<<fSDHC

sd_init_done:
	;set block size to 512 bytes
	ldid	Z, 512
	ldi	temp, 16 ;CMD16: set block size
	rcall	sd_command
	rcall	sd_clock
	
	rcall	putstr
		.db "INIT DONE",0x0a,0
	sbi	PORTD, SD_CS
	ret
	
sd_init_mmc:
	;send CMD1 until card leaves idle state
	ldi	temp, 1
	rcall	sd_command
	rcall	sd_clock
	cpi	temp, 0x01
	breq	sd_init_mmc
	cpi	temp, 0x00
	breq	sd_init_done
	;(fall through to sd_init_error)
sd_init_error:
	sbi	PORTD, SD_CS
	push temp
	rcall putstr
	.db "SD CARD INIT ERROR ",0
	pop temp
	push temp
	rcall puthex
	ldi temp, 0x0a
	rcall putc
	pop temp
	cpi	temp, 0
	brne	PC+2
	ldi	temp, 0xFF
	ret

;--------------------

sd_command:
	;send command to SD card, return read value (bit 7 set => error)
	ori	temp, 0x40
	rcall	sd_write
	mov	temp, Z4
	rcall	sd_write
	mov	temp, Z3
	rcall	sd_write
	mov	temp, Z2
	rcall	sd_write
	mov	temp, Z1
	rcall	sd_write
	ldi	temp, 0x95 ;real CRC needed for CMD0(0) and CRC8(0x122)
	rcall	sd_write
	push	count
	ldi	count, 9 ;max. 9 reads, then abort if read value is still >= 0x80
sd_command_read:
	rcall	sd_read
	cpi	temp, 0x80
	brlo	sd_command_end
	dec	count
	brne	sd_command_read
sd_command_end:
	pop	count
	ret

;--------------------

sd_write:
	;send one byte to SD card
	ldi	temp2, 8
sd_write_loop:
	cbi	PORTD, SD_MOSI
	lsl	temp
	brcc	PC+2
	sbi	PORTD, SD_MOSI
	rjmp	PC+1
	sbi	PORTD, SD_CK
	rjmp	PC+1
	cbi	PORTD, SD_CK
	dec	temp2
	brne	sd_write_loop
	ret

;--------------------

sd_read:
	;receive one byte from SD card
	sbi	PORTD, SD_MOSI
	ldi	temp2, 8
sd_read_loop:
	lsl	temp
	sbic	PIND, SD_MISO
	ori	temp, 0x01
	sbi	PORTD, SD_CK
	rjmp	PC+1
	cbi	PORTD, SD_CK
	dec	temp2
	brne	sd_read_loop
	ret

;--------------------

sd_clock:
	;send 8 clock pulses
	sbi	PORTD, SD_MOSI
	ldi	temp2, 8
sd_clock_loop:
	sbi	PORTD, SD_CK
	rjmp	PC+1
	cbi	PORTD, SD_CK
	dec	temp2
	brne	sd_clock_loop
	ret

;--------------------

str_xml_param:
	;find XML parameter RAM_String in RAM_Line and return its numeric value
	ldiw	Y, RAM_Line
str_xml_param_restart:
	movw	R1:R0, YH:YL
	ldiw	X, RAM_String
str_xml_param_loop:
	ld	temp, X+
	cpi	temp, 0
	breq	str_xml_param_found
	ld	temp2, Y+
	cpi	temp2, 0
	breq	str_xml_param_notfound
	cp	temp, temp2
	breq	str_xml_param_loop
	movw	YH:YL, R1:R0
	adiw	YH:YL, 1
	rjmp	str_xml_param_restart

str_xml_param_notfound:
	;parameter not found, return 0xFF
	ldiw	X, 0xFFFF
str_xml_param_end:
	ret
	
str_xml_param_found:
	;parameter found, return numeric value (X = value / 10, temp2 = value % 10)
	ldi	temp2, 0xFF
str_xml_param_found_loop:
	ld	temp, Y+
	cpi	temp, 0x21
	brlo	str_xml_param_end ;whitespace, non-printable char or EOL
	cpi	temp, '0'
	brlo	str_xml_param_found_loop ;ignore anything else except numbers
	cpi	temp, '9'+1
	brsh	str_xml_param_found_loop
	andi	temp, 0x0F
	cpi	temp2, 0xFF ;if value is still unset: set value = 0
	brne	str_xml_param_found_mul10
	ldiw	X, 0
	ldi	temp2, 0
str_xml_param_found_mul10:
	movw	R1:R0, XH:XL ;multiply value by 10
	lsl	R0
	rol	R1
	lsl	R0
	rol	R1
	add	XL, R0
	adc	XH, R1
	lsl	XL
	rol	XH
	add	XL, temp2 ;add previous digit
	mov	temp2, temp
	ldi	temp, 0
	adc	XH, temp
	rjmp	str_xml_param_found_loop

;--------------------

str_compare:
	;compare string at RAM[X] with string from flash
	;  PARAMS: 0x00-terminated string
	movw	R1:R0, ZH:ZL
	popw	Z
	lsl	ZL
	rol	ZH
	clt
str_compare_loop:
	lpm	temp, Z+
	cpi	temp, 0
	breq	str_compare_end
	ld	temp2, X+
	cp	temp, temp2
	breq	str_compare_loop
	set
	rjmp	str_compare_loop
str_compare_end:
	;return
	adiw	ZH:ZL, 1
	lsr	ZH
	ror	ZL
	pushw	Z
	movw	ZH:ZL, R1:R0
	ldi	temp, 0
	bld	temp, 0
	cpi	temp, 0 ;return value: Z flag (0 = equal, 1 = not equal)
	ret
	
;--------------------

str2ram:
	;copy string to RAM_String
	;  PARAMS: 0x00-terminated string (max. 10 chars including 0x00)
	movw	R1:R0, ZH:ZL
	popw	Z
	lsl	ZL
	rol	ZH
	ldiw	X, RAM_String
str2ram_loop:
	lpm	temp, Z+
	st	X+, temp
	lpm	temp2, Z+
	st	X+, temp2
	cpi	temp, 0
	breq	str2ram_end
	cpi	temp2, 0
	brne	str2ram_loop
str2ram_end:
	;return
	lsr	ZH
	ror	ZL
	pushw	Z
	movw	ZH:ZL, R1:R0
	ret
	
;--------------------

message:
	;display scrolling text message
	;  PARAMS: 0x00-terminated string (chars 0x20 to 0x7E only)
	movw	R1:R0, ZH:ZL
	popw	Z
	lsl	ZL
	rol	ZH
	
	;clear frame
	activeframe X
	ldi	count, 144
	ldi	temp, 0
message_clear:
	st	X+, temp
	dec	count
	brne	message_clear
	
message_loop:
	;get next char from flash
	lpm	temp, Z+
	cpi	temp, 0
	breq	message_end
	subi	temp, 0x20
	clr	temp2 ;multiply char # by 8 (8 bytes/char)
	lsl	temp
	rol	temp2
	lsl	temp
	rol	temp2
	lsl	temp
	rol	temp2
	pushw	Z
	ldiw	Z, charset*2
	add	ZL, temp
	adc	ZH, temp2
	ldi	count, 8
message_char_loop:
	lpm	temp, Z+
	rcall	message_scroll
	rcall	message_delay
	dec	count
	brne	message_char_loop
	popw	Z
	rjmp	message_loop
message_end:
	;continue scrolling until message has disappeared entirely (18 times)
	ldi	count, 18
message_end_wait:
	ldi	temp, 0
	rcall	message_scroll
	rcall	message_delay
	dec	count
	brne	message_end_wait
	;return
	adiw	ZH:ZL, 1
	lsr	ZH
	ror	ZL
	pushw	Z
	movw	ZH:ZL, R1:R0
	ret
	
message_scroll:
	;scroll message to the left, add new data from temp to the right
	activeframe X
	inactiveframe Y
	ldi	timeH, 8
message_scroll_line:
	adiw	XH:XL, 1 ;ignore first pixel of old frame
	ldi	timeL, 17
message_scroll_pixel:
	ld	temp2, X+
	st	Y+, temp2
	dec	timeL
	brne	message_scroll_pixel
	ldi	temp2, 0 ;0 = min. brightness (LED off)
	sbrc	temp, 0
	ldi	temp2, 7 ;7 - max. brightness (LED on)
	st	Y+, temp2
	lsr	temp
	dec	timeH
	brne	message_scroll_line
	newframe
	ret
	
message_delay:
	;short delay (~100ms)
	ldiw	time, 3
message_delay_loop:
	zero_w	time
	brne	message_delay_loop
	ret
	
.endif

;===============================================================================

;TODO: REMOVE DEBUG FUNCTIONS

puthex:
	push	temp
	swap	temp
	andi	temp, 0x0F
	ori	temp, 0x30
	cpi	temp, 0x3A
	brlo	PC+2
	subi	temp, -7
	rcall	putc
	pop	temp
	andi	temp, 0x0F
	ori	temp, 0x30
	cpi	temp, 0x3A
	brlo	PC+2
	subi	temp, -7

putc:
	sbis	UCSRA, UDRE
	rjmp	PC-1
	out	UDR, temp
	ret
	
putstr:
	;send a string to UART.
	;  PARAMS: 0x00-terminated string
	movw	R1:R0, ZH:ZL
	popw	Z
	lsl	ZL
	rol	ZH
putstr_loop:
	lpm	temp, Z+
	lpm	temp2, Z+
	cpi	temp, 0
	breq	putstr_end
	rcall	putc
	mov	temp, temp2
	cpi	temp, 0
	breq	putstr_end
	rcall	putc
	rjmp	putstr_loop
putstr_end:
	;return
	lsr	ZH
	ror	ZL
	pushw	Z
	movw	ZH:ZL, R1:R0
	ret


;===============================================================================

initframe:
;BL+ Logo
.dw 100 ;1s
.db 0xFE, 0xFE, 0x92, 0x92, 0xFE, 0x6C
.db 0x00, 0xFE, 0xFE, 0x80, 0x84, 0x84
.db 0x84, 0x1f, 0x1f, 0x04, 0x04, 0x04

frames:
;frame format:
;  * 1 word (2 bytes) duration (*10ms, little endian)
;  * 18 bytes pixel data (one byte per column, first byte = left, MSB = bottom)

;Chaosknoten
.dw 50 ;500ms
.db 0x1e, 0x21, 0x21, 0x21, 0x33, 0x2d
.db 0x3f, 0x2d, 0x2d, 0x1e, 0x0c, 0x0c
.db 0x8c, 0x54, 0xfc, 0x3c, 0x48, 0x80

;.include "movies.asm"

;end marker
.dw 0
